@@ -1,5 +1,5 @@
 /************************************************************************
- * Copyright (C) 2006-2008 Philipp Marek.
+ * Copyright (C) 2006-2009 Philipp Marek.
  *
  * This program is free software;  you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -49,12 +49,12 @@
  * url_format.
  *
  * \note
- * If there are already URLs defined, and use that command later again,
+ * If there are already URLs defined, and you use that command later again,
  * please note that as of 1.0.18 <b>the older URLs are not overwritten</b>
  * as before, but that the new URLs are \b appended to the given list!
  * If you want to start afresh, use something like
  * \code
- * echo "" | fsvs urls load
+ * true | fsvs urls load
  * \endcode
  *
  *
@@ -77,10 +77,18 @@
  *
  * To see which URLs are in use for the current WC, you can use \c dump.
  *
- * As an optional parameter you can give a format statement; \c %p , \c %n 
- * , \c %r, \c %t and \c %u are substituted by the priority, name, current 
- * revision, target revision and URL.
- * Note: That's not a real \c printf()-format; only these and a few \c \\ 
+ * As an optional parameter you can give a format statement:
+ * <table>
+ * <tr><td>\c %p<td>priority
+ * <tr><td>\c %n<td>name
+ * <tr><td>\c %r<td>current revision
+ * <tr><td>\c %t<td>target revision
+ * <tr><td>\c %R<td>readonly-flag
+ * <tr><td>\c %u<td>URL
+ * <tr><td>\c %I<td>internal number for this URL
+ * </table>
+ *
+ * \note That's not a real \c printf()-format; only these and a few \c \\ 
  * sequences are recognized.
  *
  * Example:
@@ -90,16 +98,37 @@
  *   http://svn/repos/installation/common/trunk common:50
  * \endcode 
  *
- * The default format is \c "N:%n,P:%p,D:%t,%u\\n"; for a more readable 
- * version you can use \ref glob_opt_verb "-v".
+ * The default format is \c "name:%n,prio:%p,target:%t,ro:%r,%u\\n"; for a 
+ * more readable version you can use \ref glob_opt_verb "-v".
+ *
+ *
+ * \subsection urls_load Modifying URLs
+ *
+ * You can change the various parameters of the defined URLs like this:
+ * \code
+ * # Define an URL
+ * fsvs urls name:url1,target:77,readonly:1,http://anything/...
+ * # Change values
+ * fsvs urls name:url1,target:HEAD
+ * fsvs urls readonly:0,http://anything/...
+ * fsvs urls name:url1,prio:88,target:32
+ * \endcode 
+ *
+ * \note FSVS as yet doesn't store the whole tree structures of all URLs.  
+ * So if you change the priority of an URL, and re-mix the directory trees 
+ * that way, you'll need a \ref sync-repos and some \ref revert commands.  
+ * I'd suggest to avoid this, until FSVS does handle that case better.
+ *
  * */
 
 /** \defgroup url_format Format of URLs
  * \ingroup userdoc
  *
- * FSVS can now use more than 1 URL for update. The given URLs are 
- * "overlayed" according to their priority, and they get a name (to
- * ease updating only parts).
+ * FSVS can use more than one URL; the given URLs are \e overlaid according 
+ * to their priority.
+ *
+ * For easier managing they get a name, and can optionally take a target 
+ * revision.
  *
  * Such an <i>extended URL</i> has the form
  * \code
@@ -141,9 +170,9 @@
  *
  * \section url_name Why a name?
  *
- * We need a name, so that the user can say 
- *   "<b>commit all outstanding changes to the repository at URL x</b>",
- * without having to remember the full URL.
+ * We need a name, so that the user can say <b>"commit all outstanding 
+ * changes to the repository at URL x"</b>, without having to remember the 
+ * full URL.
  * After all, this URL should already be known, as there's a list of URLs to 
  * update from.
  *
@@ -164,16 +193,16 @@
  *
  * \note In subversion you can enter \c URL\@revision - this syntax may be 
  * implemented in fsvs too. (But it has the problem, that as soon as you 
- * have a \c @ in the URL, you \b must give the target revision everytime!)
+ * have a \c @ in the URL, you \b must give the target revision every 
+ * time!)
  *
  *
  * \section url_intnum There's an additional internal number - why that?
  *
- * This internal number is not for use by the user.
+ * This internal number is not for use by the user. \n
  * It is just used to have an unique identifier for an URL, without using
  * the full string.
  *
- * \note
  * On my system the package names are on average 12.3 characters long
  * (1024 packages with 12629 bytes, including newline):
  * \code
@@ -182,11 +211,12 @@
  *
  * So if we store an \e id of the url instead of the name, we have
  * approx. 4 bytes per entry (length of strings of numbers from 1 to 1024).
- * Whereas we'd now use 12.3 characters, that's a difference of 8.3 
+ * Whereas using the needs name 12.3 characters, that's a difference of 8.3 
  * per entry.
  *
  * Multiplied with 150 000 entries we get about 1MB difference in filesize
- * of the dir-file. Not really small ...
+ * of the dir-file. Not really small ... \n
+ * And using the whole URL would inflate that much more.
  *
  * Currently we use about 92 bytes per entry. So we'd (unnecessarily) 
  * increase the size by about 10%.
@@ -201,15 +231,9 @@
  * See \ref glob_opt_urls "-u" for the specification. */
 char **url__parm_list=NULL;
 int url__parm_list_len=0,
-		url__parm_list_used;
+		url__parm_list_used=0;
 
-
-
-/** Returns whether \a url should be handled. */
-static inline int url___to_be_handled(const struct url_t *url)
-{
-  return (!url__parm_list_used) || url->to_be_handled;
-}
+int url__must_write_defs=0;
 
 
 /** -.
@@ -253,18 +277,20 @@ int url__find_by_name(const char *name, struct url_t **storage)
  * be interpreted as <i>no dirlist found</i> - which has to be allowed in 
  * some cases.
  * So this returns \c EADDRNOTAVAIL. */
-int url__find_by_url(char *url, struct url_t **storage)
+int url__find_by_url_in_list(char *url, 
+		struct url_t **list, int count,
+		struct url_t **storage)
 {
 	int status;
 	int i;
 
 
 	status=EADDRNOTAVAIL;
-	for(i=0; i<urllist_count; i++)
+	for(i=0; i<count; i++)
 	{
-		if (strcmp(urllist[i]->url, url) == 0)
+		if (strcmp(list[i]->url, url) == 0)
 		{
-			if (storage) *storage=urllist[i];
+			if (storage) *storage=list[i];
 			status=0;
 			break;
 		}
@@ -277,6 +303,12 @@ int url__find_by_url(char *url, struct url_t **storage)
 }
 
 
+/** Wrapper for url__find_by_url_in_list(). */
+int url__find_by_url(char *url, struct url_t **storage)
+{
+	return url__find_by_url_in_list(url, urllist, urllist_count, storage);
+}
+
 /** -.
  * */
 int url__find_by_intnum(int intnum, struct url_t **storage)
@@ -314,6 +346,7 @@ int url__find_by_intnum(int intnum, struct url_t **storage)
 #define HAVE_PRIO (2)
 #define HAVE_URL (4)
 #define HAVE_TARGET (8)
+#define HAVE_READONLY (16)
 /** @} */
 
 
@@ -323,7 +356,8 @@ int url__find_by_intnum(int intnum, struct url_t **storage)
  * If storage is non- \c NULL, it's \c ->name member get's a copy of the given
  * (or a deduced) name.
  *
- * In \a def_parms the parameters found are flagged - see \ref url_flags.  
+ * In \a def_parms the parameters found are flagged - see \ref url_flags; 
+ * if \a def_parms is \c NULL, an URL \b must be present.
  * */
 int url__parse(char *input, struct url_t *storage, int *def_parms)
 {
@@ -360,77 +394,129 @@ int url__parse(char *input, struct url_t *storage, int *def_parms)
 		 *             |    value
 		 *             cur
 		 *    */
-		value=strchr(cur, ':');
-		STOPIF_CODE_ERR(!value, EINVAL,
-				"!Specification '%s' is not a valid URL - ':' missing.", input);
-		value++;
+		end=cur;
+		value=NULL;
+		while (*end)
+		{
+			/* There may be multiple ':' in a single value, eg. in the URL:
+			 *   http://user:pass@host:port/.
+			 * Set value to the first occurrence. */
+			if (*end == ':' && !value)
+				value = end+1;
 
-		end=strchr(value, ',');
-		if (!end) end=value+strlen(value);
+			if (*end == ',') break;
 
-		nlen=value-cur;
-		vlen=end-value;
+			end++;
+		}	
 
-		DEBUGP("cur=%s value=%s vlen=%d nlen=%d",
-				cur, value, vlen, nlen);
+		/* Don't count the ':'. */
+		nlen = (value ? value-1 : end) - cur;
+		vlen =  value ? end - value : 0;
 
-		if (strncmp("name:", cur, nlen) == 0 ||
-				strncmp("N:", cur, nlen) == 0)
+		DEBUGP("cur=%s value=%s end=%s vlen=%d nlen=%d",
+				cur, value, end, vlen, nlen);
+
+		if (strncmp("name", cur, nlen) == 0 ||
+				strncmp("N", cur, nlen) == 0)
 		{
 			STOPIF_CODE_ERR( have_seen & HAVE_NAME, EINVAL, 
 					"!Found two names in URL '%s'; only one may be given.",
 					input);
+			if (!value) goto need_value;
+
 			/* "" == NULL == empty name? */
 			if (vlen==0)
 				DEBUGP("NULL name");
 			else if (storage) 
 			{
-				/* If we need that name again, make a copy. */
-				eurl.name=malloc(vlen+2);
-				STOPIF_ENOMEM(!eurl.name);
-				memcpy(eurl.name, value, vlen);
-				eurl.name[vlen]=0;
+				/* If we need that name again, make a copy.
+				 * We cannot simply use strdup(), because this is not 
+				 * \0-terminated. */
+				STOPIF( hlp__strnalloc(vlen, &eurl.name, value), NULL);
+
+				DEBUGP("got a name '%s' (%d bytes), going on with '%s'",
+						eurl.name, vlen, end);
+				have_seen |= HAVE_NAME;
 			}
-			DEBUGP("got a name '%s' (%d bytes), going on with '%s'",
-					eurl.name, vlen, end);
-			have_seen |= HAVE_NAME;
 		}
-		else if (strncmp("target:", cur, nlen) == 0 ||
-				strncmp("T:", cur, nlen) == 0)
+		else if (strncmp("target", cur, nlen) == 0 ||
+				strncmp("T", cur, nlen) == 0)
 		{
 			STOPIF_CODE_ERR( have_seen & HAVE_TARGET, EINVAL, 
 					"!Already got a target revision in URL '%s'.",
 					input);
+			if (!value) goto need_value;
 			STOPIF( hlp__parse_rev( value, &cp, & eurl.target_rev), NULL);
-			STOPIF_CODE_ERR( cp == value || *cp != ',', EINVAL,
+			STOPIF_CODE_ERR( cp == value || cp != end, EINVAL,
 					"The given target revision in '%s' is invalid.",
 					input);
 			DEBUGP("got target %s", hlp__rev_to_string(eurl.target_rev));
 			have_seen |= HAVE_TARGET;
 		}
-		else if (strncmp("prio:", cur, nlen) == 0 ||
-				strncmp("P:", cur, nlen) == 0)
+		else if (strncmp("prio", cur, nlen) == 0 ||
+				strncmp("P", cur, nlen) == 0)
 		{
 			STOPIF_CODE_ERR( have_seen & HAVE_PRIO, EINVAL, 
 					"!Found two priorities in URL '%s'; only one allowed.",
 					input);
+			if (!value) goto need_value;
 			eurl.priority=strtol(value, &cp, 0);
-			STOPIF_CODE_ERR( cp == value || *cp != ',', EINVAL,
-					"The given url '%s' is invalid; cannot parse the priority.",
+			STOPIF_CODE_ERR( cp == value || cp != end, EINVAL,
+					"!The given url \"%s\" is invalid; cannot parse the priority.",
 					input);
 			DEBUGP("got priority %d", eurl.priority);
 			have_seen |= HAVE_PRIO;
 		}
+		else if (strncmp("readonly", cur, nlen) == 0 ||
+				strncmp("ro", cur, nlen) == 0)
+		{
+			STOPIF_CODE_ERR( have_seen & HAVE_READONLY, EINVAL, 
+					"!Found two readonly flags in URL \"%s\"; only one allowed.",
+					input);
+			if (value) 
+			{
+				eurl.is_readonly=strtol(value, &cp, 0);
+				STOPIF_CODE_ERR( cp == value || cp != end, EINVAL,
+						"!Cannot parse the readonly flag in \"%s\".", input);
+			}
+			else
+				eurl.is_readonly=1;
+
+			have_seen |= HAVE_READONLY;
+		}
 		else
 		{
-			if (strncmp("http:", cur, nlen) == 0 ||
-					strncmp("https:", cur, nlen) == 0 ||
-					strncmp("file:", cur, nlen) == 0 ||
-					strncmp("svn:", cur, nlen) == 0 ||
-					strncmp("svn+ssh:", cur, nlen) == 0)
+			/* For URLs no abbreviation is allowed, so we check the length extra.  
+			 * An exception is "svn+", which can have arbitrary tunnels after it; 
+			 * see ~/.subversion/config for details.
+			 *
+			 * We cannot use strcmp(), as the URL has no '\0' at the given 
+			 * position.
+			 * We test for the ":", too, so that "http\0" isn't valid.
+			 * */
+			nlen++;
+			if (strncmp("svn+", cur, 4) == 0)
+			{
+					/* At least a single character after the '+'; but nlen is already 
+					 * incremented.  */
+					STOPIF_CODE_ERR(nlen <= 5, EINVAL,
+							"!No tunnel given after \"svn+\" in \"%s\".", cur);
+			}
+			else if (
+					(nlen == 4 && strncmp("svn:", cur, nlen) == 0) ||
+					(nlen == 5 &&
+					 (strncmp("http:", cur, nlen) == 0 ||
+						strncmp("file:", cur, nlen) == 0) ) ||
+					(nlen == 6 && strncmp("https:", cur, nlen) == 0))
 				DEBUGP("known protocol found");
 			else
-				STOPIF_CODE_ERR(1, EINVAL, "!The given protocol is unknown!");
+				STOPIF_CODE_ERR(1, EINVAL, 
+						"!The protocol given in \"%s\" is unknown!", cur);
+
+			/* The shortest URL is name="http:" and value="//a", or something 
+			 * like that ;-) */
+			if (!value || vlen<3 || strncmp(value, "//", 2)!=0)
+				STOPIF_CODE_ERR(1, EINVAL, "!The URL in \"%s\" is invalid.", cur);
 
 			/* Must be an URL */
 			/* We remove any / at the end of the URL (which may have resulted from
@@ -443,29 +529,40 @@ int url__parse(char *input, struct url_t *storage, int *def_parms)
 			while (vlen>3 && value[vlen-1] == '/')
 				value[--vlen] = 0;
 
-			/* We need the ":" and the "name" (protocol) too, but don't count the 
-			 * \0 at the end.  */
-			eurl.urllen=nlen + 1 + vlen - 1;
-			eurl.url=strdup(cur);
-			STOPIF_ENOMEM(!eurl.url);
+			/* We need the ':' and the "name" (protocol) too, but don't count the 
+			 * '\0' at the end.
+			 * The ':' is already counted by the nlen++ above. */
+			eurl.urllen=nlen + 0 + 1 + vlen - 1;
+			STOPIF( hlp__strdup( &eurl.url, cur), NULL);
 
 			have_seen |= HAVE_URL;
 		}
 
 		while (*end == ',') end++;
+		if (!*end) break;
 		cur=end;
 	}
 
 
-	STOPIF_CODE_ERR( !eurl.url || !*eurl.url, EINVAL,
-			"!No URL found in %s", input);
+	if (def_parms) 
+		*def_parms=have_seen;
+	else
+		STOPIF_CODE_ERR( !(have_seen & HAVE_URL), EINVAL,
+				"!No URL found in %s", input);
 
 	if (storage) *storage=eurl;
 
-	if (def_parms) *def_parms=have_seen;
+	/* Maybe not entirely correct here, because URLs might not be stored in 
+	 * the URL list. */
+	url__must_write_defs=1;
 
 ex:
 	return status;
+
+need_value:
+	STOPIF(EINVAL,
+			"!Specification '%s' is not a valid URL - ':' missing.", input);
+	goto ex;
 }
 
 
@@ -486,29 +583,44 @@ int url__insert_or_replace(char *eurl,
 {
 	int status;
 	int seen;
-	struct url_t target, *dupl, *dest;
+	struct url_t target, *dupl, *dest, *by_name;
 
 
 	status=0;
-	dupl=NULL;
 	STOPIF( url__parse(eurl, &target, &seen), NULL);
-	/* shortcut */
 
-	if (url__find_by_url(target.url, &dupl) == EADDRNOTAVAIL)
+
+	by_name=NULL;
+	/* No error checks necessary, pointer stays NULL if not found. */
+	if (seen & HAVE_NAME)
+		url__find_by_name(target.name, &by_name);
+
+	dupl=NULL;
+	/* If an URL is given, this is what is used for replacement. */
+	if (seen & HAVE_URL)
+		url__find_by_url(target.url, &dupl);
+	else
 	{
-		if (target.name)
+		/* If no URL, then try to find the name. */	
+		dupl=by_name;
+	}
+
+
+	if (!dupl)
+	{
+		if (!(seen & HAVE_URL))
+		{
+			STOPIF( EINVAL,
+					!(seen & HAVE_NAME) ? 
+					"!No URL was given in \"%s\"." :
+					"!Cannot find the name given in \"%s\", so cannot modify an URL.",
+					eurl);
+		}
+		if (seen & HAVE_NAME)
 		{
 			/* The names must be unique. */
-			status=url__find_by_name(target.name, NULL);
-			if (status == EADDRNOTAVAIL)
-			{
-				/* Good, just what we want to hear. */
-			}
-			else if (status == 0)
-				STOPIF( EADDRINUSE, 
-						"!There's already an url named '%s'", target.name);
-			else
-				STOPIF( status, "Cannot look for url named '%s'", target.name);
+			STOPIF_CODE_ERR( by_name, EADDRINUSE, 
+					"!There's already an url named \"%s\"", target.name);
 
 			/* If we didn't find it, it's ok. */
 			status=0;
@@ -531,6 +643,8 @@ int url__insert_or_replace(char *eurl,
 			dupl->target_rev = target.target_rev;
 		if (seen & HAVE_PRIO)
 			dupl->priority = target.priority;
+		if (seen & HAVE_READONLY)
+			dupl->is_readonly = target.is_readonly;
 		if (seen & HAVE_NAME)
 			dupl->name = target.name;
 		/* The URL is the same, so the length is the same, and the internal 
@@ -645,11 +759,9 @@ int url__allocate(int reserve_space)
 
 	status=0;
 	/* We put a terminating NULL pointer at the end. */
-	urllist=realloc(urllist, 
-			sizeof(*urllist) * (urllist_count+1+reserve_space));
-	STOPIF_ENOMEM(!urllist);
-	url_mem=calloc(sizeof(*url_mem), reserve_space);
-	STOPIF_ENOMEM(!url_mem);
+	STOPIF( hlp__realloc( &urllist, 
+			sizeof(*urllist) * (urllist_count+1+reserve_space)), NULL);
+	STOPIF( hlp__calloc( &url_mem, sizeof(*url_mem), reserve_space), NULL);
 
 	/* store url pointers */
 	for(i=0; i<reserve_space; i++)
@@ -663,23 +775,13 @@ ex:
 }
 
 
-/** Comparing two URLs.
- *
- * They get sorted by \a priority ascending (lower numbers, so higher
- * priority, first), then by \a url ascending (sort URLs alphabetically). 
- *
- * This is necessary, as on update we walk the \a urllist in order, to
- * have lower priority entries appearing when higher priority entries are
- * removed. */
-static int url___sorter(const void *a, const void *b)
+/** -. */
+int url__indir_sorter(const void *a, const void *b)
 {
 	struct url_t *u1=*(struct url_t **)a,
 							 *u2=*(struct url_t **)b;
 
-	if (u1->priority == u2->priority)
-		return strcmp(u1->url, u2->url);
-	else
-		return u1->priority - u2->priority;
+	return url__sorter(u1, u2);
 }
 
 
@@ -692,15 +794,24 @@ static int url___sorter(const void *a, const void *b)
  * If no \c dir file is found, \c ENOENT is returned without an error 
  * message.
  *
+ * We read two sources; the one in \c FSVS_CONF normally holds
+ * the URL, the target revision, the priority, the name, the internal 
+ * number, and the readonly flag.
+ *
+ * The current revision is read from \c FSVS_WAA.
+ *
  * \see waa_files. */
 int url__load_list(char *dir, int reserve_space)
 {
 	int status, fh, l, i;
-	struct stat64 st;
+	struct stat st;
 	char *urllist_mem;
 	int inum, cnt, new_count;
 	svn_revnum_t rev;
+	int intnum;
 	struct url_t *target;
+	FILE *rev_in;
+	char *buffer;
 
 
 	fh=-1;
@@ -718,12 +829,11 @@ int url__load_list(char *dir, int reserve_space)
 
 	STOPIF_CODE_ERR(status, status, "Cannot read URL list");
 
-	STOPIF_CODE_ERR( fstat64(fh, &st) == -1, errno,
+	STOPIF_CODE_ERR( fstat(fh, &st) == -1, errno,
 			"fstat() of url-list");
 
 	/* add 1 byte to ensure \0 */
-	urllist_mem=malloc(st.st_size+1);
-	STOPIF_ENOMEM(!urllist_mem);
+	STOPIF( hlp__alloc( &urllist_mem, st.st_size+1), NULL);
 
 	status=read(fh, urllist_mem, st.st_size);
 	STOPIF_CODE_ERR( status != st.st_size, errno, 
@@ -772,12 +882,58 @@ int url__load_list(char *dir, int reserve_space)
 		l++;
 	}
 
+	STOPIF_CODE_ERR( close(fh) == -1, errno, "closing the url-list");
+	fh=-1;
+
+
+	/* Read the current revisions from the WAA definition.
+	 * If we got data before, we need this here too. */
+	/* Exception for 1.1.18: Upgrade from 1.1.17. A non-existing file is 
+	 * allowed, but will convert the data next time. */
+	status=waa__open_byext(dir, WAA__URL_REVS, WAA__READ, &fh);
+	if (status==ENOENT)
+	{
+		DEBUGP("No file; upgrading?");
+		status=0;
+	}
+	else
+	{
+		/* Read the associated revisions. */
+		rev_in=fdopen(fh, "r");
+		while (1)
+		{
+			status=hlp__string_from_filep(rev_in, &buffer, NULL, 0);
+			if (status == EOF) 
+			{
+				status=0;
+				break;
+			}
+			STOPIF( status, "Failed to read copyfrom source");
+
+			STOPIF_CODE_ERR( sscanf(buffer, "%d %lu 0 0 0 0\n", 
+						&intnum, &rev) != 2, EINVAL,
+					"Error parsing line \"%s\" from %s", buffer, WAA__URL_REVS);
+
+			STOPIF( url__find_by_intnum(intnum, &target), 
+					"URL number %d read from %s not found",
+					intnum, WAA__URL_REVS);
+
+			target->current_rev=rev;
+		}
+		STOPIF_CODE_ERR( fclose(rev_in)==-1, errno, 
+			"error closing %s", WAA__URL_REVS);
+		fh=-1;
+	}
+
+
 	/* Sort list by priority */
-	qsort(urllist, urllist_count, sizeof(*urllist), url___sorter);
+	qsort(urllist, urllist_count, sizeof(*urllist), url__indir_sorter);
+
+	/* No writing would be necessary if nothing gets changed. */
+	url__must_write_defs=0;
 
 ex:
 	/* urllist_mem must not be freed - our url-strings still live there! */
-
 	if (fh!=-1) 
 	{
 		l=close(fh);
@@ -809,47 +965,76 @@ ex:
 
 	
 /** -.
+ * The data is written in two different locations.
+ *
+ * The internal number was chosen as combining key, because the URL might 
+ * include strange characters, and there might not be a name.
+ *
  * \todo is 1024 bytes always enough? Maybe there's an RFC. Make that
  * dynamically - look how much we'll need. */
 int url__output_list(void)
 {
-	int status, i, fh, l;
+	int status, i, fh, l, fh_revs;
 	char buffer[1024];
 	struct url_t *url;
 
 
 	fh=-1;
+	fh_revs=-1;
 
 	STOPIF( url___set_internal_nums(), 
 			"Setting the internal numbers failed.");
 
-	STOPIF( waa__open_byext(NULL, WAA__URLLIST_EXT, WAA__WRITE, &fh), NULL);
+	if (url__must_write_defs)
+		STOPIF( waa__open_byext(NULL, WAA__URLLIST_EXT, WAA__WRITE, &fh), NULL);
+
+	STOPIF( waa__open_byext(NULL, WAA__URL_REVS, WAA__WRITE, &fh_revs), NULL);
 	for(i=0; i<urllist_count; i++)
 	{
 		url=urllist[i];
 
+		if (url->target_rev == 0 && url->current_rev == 0)
+			continue;
+
+		if (fh != -1)
+		{
+			l=snprintf(buffer, sizeof(buffer),
+					"%d %d T:%ld,N:%s,P:%d,ro:%u,%s",
+					url->internal_number,
+					0, /* Previously the the current revision. */
+					url->target_rev,
+					url->name ? url->name : "",
+					url->priority,
+					url->is_readonly,
+					url->url);
+
+			STOPIF_CODE_ERR( l > sizeof(buffer)-4, E2BIG, 
+					"You've got too long URLs; I'd need %d bytes. Sorry.", l);
+
+			/* include the \0 */
+			l++;
+			/** \todo: writev */
+			STOPIF_CODE_ERR( write(fh, buffer, l) != l, errno,
+					"Error writing the URL list");
+			STOPIF_CODE_ERR( write(fh, "\n", 1) != 1, errno,
+					"Error writing the URL list delimiter");
+			DEBUGP("writing line %s", buffer);
+		}
+
+
+		/* A few extra fields, to store some information later. */
 		l=snprintf(buffer, sizeof(buffer),
-				"%d %ld T:%ld,N:%s,P:%d,%s",
+				"%d %ld 0 0 0 0\n",
 				url->internal_number,
-				url->current_rev,
-				url->target_rev,
-				url->name ? url->name : "",
-				url->priority,
-				url->url);
-
-		STOPIF_CODE_ERR( l > sizeof(buffer)-4, E2BIG, 
-				"You've got too long URLs; I'd need %d bytes. Sorry.", l);
-
-		/* include the \0 */
-		l++;
-		/** \todo: writev */
-		STOPIF_CODE_ERR( write(fh, buffer, l) != l, errno,
+				url->current_rev);
+		/* This can never happen, apart from being caused by radiation. */
+		BUG_ON( l > sizeof(buffer)-4);
+		STOPIF_CODE_ERR( write(fh_revs, buffer, l) != l, errno,
 				"Error writing the URL list");
-		STOPIF_CODE_ERR( write(fh, "\n", 1) != 1, errno,
-				"Error writing the URL list delimiter");
-		DEBUGP("writing line %s", buffer);
 	}
 
+	url__must_write_defs=0;
+
 ex:
 	if (fh != -1)
 	{
@@ -858,17 +1043,40 @@ ex:
 		STOPIF(i, "Error closing the URL list");
 	}
 
+	if (fh_revs != -1)
+	{
+		i=waa__close(fh_revs, status);
+		fh_revs=-1;
+		STOPIF(i, "Error closing the revisions list");
+	}
+
 	return status;
 }
 
 
 /** -.
+ *
+ * If \a missing_dirs is not \c NULL, this function returns in \c 
+ * *missing_dirs a \b copied string with the missing path components from 
+ * \c current_url->url (which should be freed later).
+ *
+ * \note The session is then registered at the \b existing part, so all 
+ * accesses must include this relative part!
+ *
+ * If the URL is ok, a \c NULL is returned (not a pointer to a \c \\0 ).
+ *
+ * This is needed for the \c mkdir_base option; we cannot create the 
+ * hierarchy here, because we need a commit editor for that, but in 
+ * ci__directory() we cannot use a session based on an non-existing URL.
  * */
-int url__open_session(svn_ra_session_t **session)
+int url__open_session(svn_ra_session_t **session, char **missing_dirs)
 {
 	int status;
 	svn_error_t *status_svn;
 	apr_hash_t *cfg;
+	char *buffer, *cp;
+	int exists;
+	svn_revnum_t head;
 
 
 	status=0;
@@ -881,34 +1089,123 @@ int url__open_session(svn_ra_session_t **session)
 
 	STOPIF( hlp__get_svn_config(&cfg), NULL);
 
-	/** Try svn_ra_reparent() */
-	if (!current_url->session)
-	{
-		STOPIF_SVNERR_TEXT( svn_ra_open,
-				(& current_url->session, current_url->url,
+
+	if (current_url->session) goto ex;
+
+
+	/* We wouldn't need to allocate this memory if the URL was ok; but we 
+	 * don't know that here, and it doesn't hurt that much. */
+	STOPIF( hlp__strnalloc(current_url->urllen, 
+				&buffer, current_url->url), NULL);
+	cp=buffer+current_url->urllen;
+	BUG_ON(*cp);
+
+	STOPIF_SVNERR_TEXT( svn_ra_open,
+				(& current_url->session, buffer,
 				 &cb__cb_table, NULL,  /* cbtable, cbbaton, */
 				 cfg,	/* config hash */
 				 current_url->pool),
 				"svn_ra_open(\"%s\")", current_url->url);
+	head=SVN_INVALID_REVNUM;
+  STOPIF( url__canonical_rev( current_url, &head), NULL);
 
-		if (session) 
-			*session = current_url->session;
+	DEBUGP("Trying url %s@%ld", buffer, head);
+	while (1)
+	{
+		/* Is the caller interested in this check? If not, then just return. */
+		if (!missing_dirs) break;
 
-		DEBUGP("Opened url %s", current_url->url);
+
+		/* Test whether the base directory exists; we need some lightweight 
+		 * mechanism to detect that.
+		 * Sadly we don't get a result when we open the session. */
+		/* That's not entirely correct.
+		 * In the time between this test and the commit running someone could 
+		 * create or remove the base path; then we would have tested against 
+		 * the wrong revision, and might fail nonetheless. */
+		STOPIF( cb__does_path_exist(current_url->session, "", head, 
+					&exists, current_url->pool), NULL);
+		if (exists) break;
+
+
+		/* Doesn't exist. Try with the last part removed. */
+		/* We don't do URLs with less than a few characters. */
+		while (cp > buffer+4 && *cp != '/') cp--;
+
+		/* If we're before the hostname, signified by a "//", we abort. */
+		STOPIF_CODE_EPIPE(cp[-1] == '/', EINVAL,
+				"!Unsuccessfull svn_ra_stat() on every try for URL \"%s\".",
+				current_url->url);
+
+		/* We're at a slash, and try with a shortened URL. */
+		*cp=0;
+
+		DEBUGP("Reparent to %s", buffer);
+		STOPIF_SVNERR( svn_ra_reparent,
+				(current_url->session, buffer, current_url->pool));
 	}
 
+
+	/* See whether the original URL is valid. */
+	if (missing_dirs) 
+	{
+		if (buffer + current_url->urllen == cp)
+		{
+			*missing_dirs=NULL;
+			IF_FREE(buffer);
+		}
+		else
+		{
+			/* Return just the missing parts:
+			 *
+			 *   url:    http://aaa/11/22/33/44
+			 *   buffer: http://aaa/11/22
+			 *   return: 33/44
+			 *
+			 * We return the characters that were cut off, without the '/'. */
+			strcpy(buffer, current_url->url + 1 + (cp - buffer));
+
+			DEBUGP("returning missing=%s", buffer);
+			*missing_dirs=buffer;
+		}
+	}
+	else IF_FREE(buffer);
+
+
+	if (session) 
+		*session = current_url->session;
+
 ex:
 	return status;
 }
 
 
+/** -.
+ * */
+int url__close_session(struct url_t *cur)
+{
+	/* There's no svn_ra_close() or suchlike.
+	 * I hope it gets closed by freeing it's pool. */
+	if (cur->pool)
+	{
+		DEBUGP("closing session and pool for %s", cur->url);
+
+		BUG_ON(cur->pool == NULL && cur->session != NULL);
+		apr_pool_destroy(cur->pool);
+		cur->session=NULL;
+		cur->pool=NULL;
+	}
+
+	return 0;
+}
+
+
 /** -.
  * */
 int url__close_sessions(void)
 {
   int status;
 	int i;
-	struct url_t *cur;
 
 	status=0;
 
@@ -916,22 +1213,9 @@ int url__close_sessions(void)
 	url__parm_list_len=url__parm_list_used=0;
 
 	for(i=0; i<urllist_count; i++)
-	{
-		cur=urllist[i];
-
-		/* There's no svn_ra_close() or suchlike.
-		 * I hope it gets closed by freeing it's pool. */
-		if (cur->pool)
-		{
-			DEBUGP("closing session and pool for %s", cur->url);
-
-			BUG_ON(cur->pool == NULL && cur->session != NULL);
-			apr_pool_destroy(cur->pool);
-			cur->session=NULL;
-			cur->pool=NULL;
-		}
-	}
+		STOPIF( url__close_session( urllist[i] ), NULL);
 
+ex:
 	return status;
 }
 
@@ -958,9 +1242,10 @@ int url___dump(char *format)
 
 
 	if (!format) 
-		format= opt_verbose>0 ? 
-			"%u\\n\tname: \"%n\"; priority: %p; current revision: %r; target: %t;\\n":
-			"N:%n,P:%p,%u\\n";
+		format= opt__is_verbose()>0 ? 
+			"%u\\n\tname: \"%n\"; priority: %p; current revision: %r; "
+			"target: %t; readonly:%R\\n" :
+			"name:%n,prio:%p,target:%t,ro:%R,%u\\n";
 
 	status=0;
 	for(i=0; i < urllist_count; i++)
@@ -1024,6 +1309,14 @@ int url___dump(char *format)
 										hlp__rev_to_string(url->current_rev), 
 										output), NULL);
 							break;
+						case 'R':
+							STOPIF_CODE_EPIPE( fprintf(output, "%u", 
+										url->is_readonly), NULL);
+							break;
+						case 'I':
+							STOPIF_CODE_EPIPE( fprintf(output, "%u", 
+										url->internal_number), NULL);
+							break;
 						case 'p':
 							STOPIF_CODE_EPIPE( fprintf(output, "%u", 
 										url->priority), NULL);
@@ -1059,25 +1352,24 @@ ex:
 
 /** -.
  * The space for the output is allocated, and must not be freed. */
-int url__full_url(struct estat *sts, char *path, char **url)
+int url__other_full_url(struct estat *sts, struct url_t *url, char **output)
 {	
 	static const char none[]="(none)";
 	static struct cache_t *cache=NULL;
 	int status, len;
-	char *data;
+	char *data, *path;
 
 	status=0;
 
-	if (sts->url)
+	if (url)
 	{
-		len=sts->url->urllen + 1 + sts->path_len+1;
+		STOPIF( ops__build_path( &path, sts), NULL);
+		len=url->urllen + 1 + sts->path_len+1;
 		STOPIF( cch__new_cache(&cache, 4), NULL);
 
 		STOPIF( cch__add(cache, 0, NULL, len, &data), NULL);
-    strcpy( data, sts->url->url);
+    strcpy( data, url->url);
 
-		if (!path)
-		  STOPIF( ops__build_path( &path, sts), NULL);
 		if (path[0]=='.' && path[1]==0) 
 		{
 			/* Nothing to be done; just the base URL. */
@@ -1087,20 +1379,33 @@ int url__full_url(struct estat *sts, char *path, char **url)
 			/* Remove ./ at start. */
 			if (path[0]=='.' && path[1]==PATH_SEPARATOR) path += 2;
 
-			data[sts->url->urllen]='/';
-			strcpy( data+sts->url->urllen+1, path);
+			data[url->urllen]='/';
+			strcpy( data+url->urllen+1, path);
 		}
 
-		*url=data;
+		*output=data;
 	}
 	else
-		*url=(char*)none;
+		*output=(char*)none;
 
 ex:
 	return status;
 }
 
 
+/** -. */
+int url__full_url(struct estat *sts, char **url)
+{	
+	int status;
+
+	STOPIF( url__other_full_url(sts, sts->url, url), NULL);
+
+ex:
+	return status;
+}
+
+
+
 /** -. */
 int url__find(char *url, struct url_t **output)
 {
@@ -1132,12 +1437,17 @@ int url__work(struct estat *root UNUSED, int argc, char *argv[])
 	char *cp;
 	int have_space;
 	struct url_t *target;
+	struct url_t *tmp;
+	struct url_t **old_urllist;
+	int old_urllist_count;
 
 
 	dir=NULL;
 	fh=-1;
 
 	STOPIF( waa__given_or_current_wd(NULL, &dir), NULL );
+	/* The current directory is the WC root. */
+	STOPIF( waa__set_working_copy(dir), NULL);
 
 	/* If there's \b no parameter given, we default to dump.
 	 * - Use goto?
@@ -1145,6 +1455,38 @@ int url__work(struct estat *root UNUSED, int argc, char *argv[])
 	 * - Test for argc? That's what we'll do. */
 	if (argc>0 && strcmp(argv[0], parm_load) == 0)
 	{
+		/* In case the user had some URLs already defined and "load"s another 
+		 * list, he would loose all URL internal numbers, so that a 
+		 * "sync-repos" would be necessary.
+		 *
+		 * To avoid that we read the existing URLs (but ignore any errors, in 
+		 * case the URLs are loaded again because the file is damaged). */
+		status=url__load_list(NULL, argc+1);
+		if (!status || status == ENOENT)
+		{
+			/* all right, just ignore that. */
+		}
+		else
+		{
+			/* Other errors are at least shown. */
+			STOPIF_CODE_ERR_GOTO( 1, status, ignore_err,
+					"!Got an error reading the old URL list, so the internal URL mappings\n"
+					"cannot be kept; a \"sync-repos\" might be necessary.");
+ignore_err:
+			;
+		}
+
+		/* Don't remember the old values. */
+		old_urllist_count=urllist_count;
+		old_urllist=urllist;
+		urllist=NULL;
+		urllist_count=0;
+
+		/* Surely write the list again. */
+		url__must_write_defs=1;
+
+		status=0;
+
 		/* Load URLs. We do not know how many we'll get, possibly we'll have
 		 * to allocate more memory. */
 		i=0;
@@ -1157,7 +1499,7 @@ int url__work(struct estat *root UNUSED, int argc, char *argv[])
 				STOPIF( url__allocate(have_space), NULL);
 			}
 
-			status=hlp__string_from_filep(stdin, &cp, 1);
+			status=hlp__string_from_filep(stdin, &cp, NULL, SFF_WHITESPACE);
 			if (status == EOF) break;
 
 			DEBUGP("parsing %s into %d", cp, urllist_count);
@@ -1169,9 +1511,16 @@ int url__work(struct estat *root UNUSED, int argc, char *argv[])
 				i++;
 			}
 			target->current_rev=0;
+
+			/* Try to restore the internal number. */
+			if (url__find_by_url_in_list(target->url, 
+						old_urllist, old_urllist_count, &tmp) == 0)
+				target->internal_number = tmp->internal_number;
 		}
 
-		if (opt_verbose>=0)
+		IF_FREE(old_urllist);
+
+		if (opt__is_verbose() >= 0)
 			printf("%d URL%s loaded.\n", i, i==1 ? "" : "s");
 	}
 	else
@@ -1184,9 +1533,12 @@ int url__work(struct estat *root UNUSED, int argc, char *argv[])
 		else
 			STOPIF_CODE_ERR( status, status, NULL);
 
-
+		/* Needs status still set from url__load_list()! */
 		if (argc == 0 || strcmp(argv[0], parm_dump) == 0)
 		{
+			STOPIF_CODE_ERR( status==ENOENT, ENOENT,
+				 "!No URLs defined for \"%s\".", dir);
+
 			/* Dump */
 			STOPIF( url___dump(argc ? argv[1] : NULL), NULL);
 			goto ex;
@@ -1205,11 +1557,10 @@ int url__work(struct estat *root UNUSED, int argc, char *argv[])
 		}
 	} /* if load_from_stdin */
 
+	STOPIF( waa__create_working_copy(dir), NULL);
 
+	/* Write the URL list */
 	STOPIF( url__output_list(), NULL);
-	/* Make an informational link to the base directory. */
-	/* Should we ignore errors? */
-	STOPIF( waa__make_info_link(dir, "_base", dir), NULL);
 
 ex:
 	return status;
@@ -1284,9 +1635,8 @@ int url__store_url_name(char *parm)
 	if (url__parm_list_used+2 >= url__parm_list_len)
 	{
 		url__parm_list_len= url__parm_list_len ? url__parm_list_len*2 : 8;
-		url__parm_list=realloc(url__parm_list, 
-				url__parm_list_len*sizeof(*url__parm_list));
-		STOPIF_ENOMEM(!url__parm_list);
+		STOPIF( hlp__realloc( &url__parm_list, 
+				url__parm_list_len*sizeof(*url__parm_list)), NULL);
 	}
 
 	url__parm_list[url__parm_list_used++] = parm;
@@ -1338,14 +1688,17 @@ ex:
 /** -.
  * Returns 0 as long as there's an URL to process; \c current_url is set, 
  * and opened. In \a target_rev the target revision (as per default of this 
- * URL, or as given by the user) is returned.
+ * URL, or as given by the user) is returned. \n
+ * If \c current_url is not NULL upon entry the connection to this URL is 
+ * closed, and its memory freed.
  *
  * If called with \a target_rev \c NULL, the internal index is reset, and 
  * no URL initialization is done.
  *
  * At the end of the list \c EOF is given.
  * */
-int url__iterator(svn_revnum_t *target_rev) // , svn_revnum_t *cur_rev)
+int url__iterator2(svn_revnum_t *target_rev, int only_if_count,
+		char **missing)
 {
 	int status;
 	static int last_index=-1;
@@ -1359,6 +1712,7 @@ int url__iterator(svn_revnum_t *target_rev) // , svn_revnum_t *cur_rev)
 		goto ex;
 	}
 
+
 	while (1)
 	{
 		last_index++;
@@ -1371,10 +1725,22 @@ int url__iterator(svn_revnum_t *target_rev) // , svn_revnum_t *cur_rev)
 		}
 
 		current_url=urllist[last_index];
-		if (url___to_be_handled(current_url)) break;
+
+		if (only_if_count)
+		{
+			if (!current_url->entry_list_count)
+			{
+				DEBUGP("No changes for url %s.", current_url->url);
+				continue;
+			}
+			DEBUGP("%d changes for url %s.", 
+					current_url->entry_list_count, current_url->url);
+		}
+
+		if (url__to_be_handled(current_url)) break;
 	}
 
-	STOPIF( url__open_session(NULL), NULL);
+	STOPIF( url__open_session(NULL, missing), NULL);
 
 
 	if (current_url->current_target_override)
@@ -1383,7 +1749,8 @@ int url__iterator(svn_revnum_t *target_rev) // , svn_revnum_t *cur_rev)
 		rev=opt_target_revision;
 	else
 		rev=current_url->target_rev;
-	DEBUGP("doing URL %s @ %llu", current_url->url, (t_ull)rev);
+	DEBUGP("doing URL %s @ %s", current_url->url, 
+			hlp__rev_to_string(rev));
 
 	STOPIF( url__canonical_rev(current_url, &rev), NULL);
 	*target_rev = rev;
